// -*- C++ -*-

#pragma once

#include "gpu-utils.h"

/**
 \addtogroup debug debugging
    For debugging, various methods are provided. The gpu_ostream streaming class can be used to provide output from within a kernel. The gpu_assert macro can be placed in kernels for validity checks. The checknan data type does extensive automatic error checks on memory access.
 
 */


namespace goopax {
namespace impl {
struct dumpinfo_base
{
    using infotype_t = unsigned int;

    ostream&     os;
    virtual void out(const infotype_t*) const = 0;
    inline dumpinfo_base(ostream& os0)
        : os(os0)
    {}
    virtual ~dumpinfo_base() {}
};

namespace interface {
extern "C" {
GOOPAX_DLL_EXPORT(4.0.0) errno_type goopax_4E_dumpstream_write(gpu_writer_base_impl** vals, size_t valnum, const char* comment, dumpinfo_base* di, void (*)(dumpinfo_base*, const dumpinfo_base::infotype_t*), void (*)(dumpinfo_base*), bool) noexcept;
GOOPAX_DLL_EXPORT(4.0.0) unsigned int goopax_4_dumpstream_getscope() noexcept;
}
}


template<class T>
struct dumpinfo;

template<>
struct dumpinfo<string> : public dumpinfo_base
{
    string             str;
    const unsigned int scope_ID;
    void               out(const infotype_t*) const
    {
        os << str;
    }
    dumpinfo(const string& str0, ostream& os)
        : dumpinfo_base(os)
        , str(str0)
        , scope_ID(interface::goopax_4_dumpstream_getscope())
    {}
};

template<class FUNC>
struct dumpinfo_manip : public dumpinfo_base
{
    FUNC __pf;

    void out(const infotype_t*) const
    {
        os << *__pf;
    }
    dumpinfo_manip(FUNC __pf0, ostream& os)
        : dumpinfo_base(os)
        , __pf(__pf0)
    {}
};

template<>
struct dumpinfo<double> : public dumpinfo_base
{
    void out(const infotype_t* data) const
    {
        os << reinterpret<double>(array<unsigned int, 2>({ { data[0], data[1] } }));
    }
    dumpinfo(ostream& os)
        : dumpinfo_base(os)
    {}
};

template<class T>
struct dumpinfo : public dumpinfo_base
{
    void out(const infotype_t* data) const
    {
        static const size_t  NE     = (gettype<T>::size / gettype<gpu_uint>::size >= 1 ? gettype<T>::size / gettype<gpu_uint>::size : 1);
        static const size_t  blowup = (gettype<gpu_uint>::size / gettype<T>::size >= 1 ? gettype<gpu_uint>::size / gettype<T>::size : 1);
        vector<unsigned int> use(data, data + NE);
        os << reinterpret<array<T, blowup>>(use)[0];
    }
    dumpinfo(ostream& os)
        : dumpinfo_base(os)
    {}
};


/// \ingroup debug
/// Streaming class that can be used from within GPU kernels. The usage is similar to std::ostream
/// The output is written to the std::ostream object specified in the constructor. The output may be delayed until wait() has been called for the kernel/device
/// The order in which concurrent threads write is undefined.
/// The output of any thread is only interrupted by other threads when 'endl' or 'flush' is used. If the output should not be interrupted, use "\n" instead of endl.
class gpu_ostream
{
    using infotype_t = dumpinfo_base::infotype_t;
    ostream& os;

    static void dumpinfo_caller(dumpinfo_base* di, const infotype_t* data)
    {
        di->out(data);
    }
    static void dumpinfo_deleter(dumpinfo_base* di)
    {
        delete di;
    }

    void put(dumpinfo_base* ne, const vector<gpu_uint>& vals, const string& comment, bool flush = false)
    {
        vector<gpu_writer_base_impl*> vimp(vals.size());
        for (size_t k = 0; k < vals.size(); ++k)
        {
            vimp[k] = vals[k].impl.get();
        }
        check_errno(interface::goopax_4E_dumpstream_write((vimp.empty() ? nullptr : &vimp[0]), vimp.size(), comment.c_str(), ne, dumpinfo_caller, dumpinfo_deleter, flush));
    }

    public:
    /// output string
    friend gpu_ostream& operator<<(gpu_ostream& out, const string& s)
    {
        out.put(new dumpinfo<string>(s, out.os), {}, "DUMP<string> \"" + s + "\"");
        return out;
    }
    /// output gpu value
    template<class T>
    friend gpu_ostream& operator<<(gpu_ostream& out, const gpu_type<T>& val)
    {
        static const size_t NE     = (gettype<T>::size / gettype<gpu_uint>::size >= 1 ? gettype<T>::size / gettype<gpu_uint>::size : 1);
        static const size_t blowup = (gettype<gpu_uint>::size / gettype<T>::size >= 1 ? gettype<gpu_uint>::size / gettype<T>::size : 1);


#if GOOPAX_DEBUG
        array<gpu_type<T>, blowup> srcvec;
        srcvec.fill(0);
#else
        array<gpu_type<T>, blowup> srcvec;
        for (auto& s : srcvec)
            s = 0;
#endif
        srcvec[0]               = val;
        array<gpu_uint, NE> use = reinterpret<array<gpu_uint, NE>>(srcvec);

        out.put(new dumpinfo<typename gettype<T>::cpu>(out.os),
                vector<gpu_uint>(use.begin(), use.end()),
                "dump value<gpu_float>");
        return out;
    }
    /// output cpu value
    template<typename T, typename = typename enable_if<is_fundamental<T>::value || is_half<T>::value>::type>
    friend gpu_ostream& operator<<(gpu_ostream& s, T a)
    {
        return s << gpu_type<T>(a);
    }


    /// output STL vector/array
    template<class V>
    friend
        typename enable_if<std::is_base_of<std::random_access_iterator_tag, typename V::iterator>::value, gpu_ostream>::type&
        operator<<(gpu_ostream& out, const V& v)
    {
        out << "(";
        for (size_t k = 0; k < v.size(); ++k)
        {
            if (k != 0) out << ",";
            out << v[k];
        }
        out << ")";
        return out;
    }
    /// output C-style string
    friend gpu_ostream& operator<<(gpu_ostream& out, const char* s)
    {
        return out << string(s);
    }

    /// output endl/flush
    friend gpu_ostream& operator<<(gpu_ostream& s, ostream& (*__pf)(ostream&))
    {

        s.put(new dumpinfo_manip<ostream& (*)(ostream&)>(__pf, s.os),
              {},
              "ostream << iomanip",
              true);
        return s;
    }
    /// modify stream, e.g. by hex or dec modifyers
    friend gpu_ostream& operator<<(gpu_ostream& s, decltype(&hex) h)
    {
        s.put(new dumpinfo_manip<decltype(&hex)>(h, s.os),
              {},
              "ostream << ios_manipulator");
        return s;
    }
    /// output cpu debug value
    template<class T, class TA>
    friend inline gpu_ostream& operator<<(gpu_ostream& out, const checknan<T, TA>& val)
    {
        return out << typename gettype<T>::gpu(val);
    }

    /// Constructor.
    /// \param os0 std::ostream object to which the output is passed.
    gpu_ostream(ostream& os0)
        : os(os0)
    {}
    ~gpu_ostream()                  = default;
    gpu_ostream(const gpu_ostream&) = delete;
    gpu_ostream& operator=(const gpu_ostream&) = delete;
};
}
using impl::gpu_ostream;
}
